/**************************************************************************
*  Copyright (c) 2018 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  29.04.2018  mifi  First Version.
**************************************************************************/
#define __NUTNET_LPC17XX_40XX_ETH_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include "tal.h"
#include "ipstack.h"
#include "eth_phy_driver.h"

#include "netinet/if_ether.h"
#include "net/ether.h"

#include "enet_24xx_17xx_40xx.h"

/*=======================================================================*/
/*  Extern                                                               */
/*=======================================================================*/

/* Configure the GPIO ports for ethernet pins */
extern ETH_PHY_DRIVER *BoardETHConfig (void);

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

#define NETIF_EVENT_RX           0x0001
#define NETIF_EVENT_TX           0x0002
#define NETIF_EVENT_LINK_UP      0x0004   /* Not used */
#define NETIF_EVENT_LINK_DOWN    0x0008   /* Not used */
#define NETIF_EVENT_RX_OVERRUN   0x0010
#define NETIF_EVENT_TX_UNDERRUN  0x0020

#define NETIF_EVENT_WAIT_ALL  (NETIF_EVENT_RX         | \
                               NETIF_EVENT_TX         | \
                               NETIF_EVENT_LINK_UP    | \
                               NETIF_EVENT_LINK_DOWN  | \
                               NETIF_EVENT_RX_OVERRUN | \
                               NETIF_EVENT_TX_UNDERRUN)


/* NIC control block */
typedef struct
{
   uint8_t         ni_mac[6];    /* MAC Address */    
   ETH_PHY_DRIVER *ni_phy;   
} NICINFO;

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

/*
 * This is the stack for the ReceiverTask
 */
static OS_STACK (ReceiverStack, TASK_IP_RX_STK_SIZE);
static os_tcb_t   tcb_task_rx;

static NICINFO    dcb_eth0;

static os_event_t NetIfEvent;

static uint8_t OldLinkStatus = PHY_LINK_STATUS_NO_LINK; 

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/**
 * Timer callback function for checking the Ethernet Link.
 *
 * @param arg contain the PHYDriver
 */
static void CheckLink (ETH_PHY_DRIVER *pPHY )
{
   uint8_t        LinkStatus;
   
   /* Get actual Link status */
   LinkStatus = pPHY->GetLinkStatus(pPHY);
   
   /* Check if the status has changed */
   if (LinkStatus != OldLinkStatus)
   {
      /* Check for DOWN/UP */
      if (PHY_LINK_STATUS_NO_LINK == LinkStatus)
      {
         /* Do nothing */
      }
      else
      {
         /* Update MAC */
         enet_UpdateLink(LinkStatus);
      }
      
      IPIfaceLinkCallback(LinkStatus);
   } /* end if (LinkStatus != OldLinkStatus) */
   
   OldLinkStatus = LinkStatus;      
} /* CheckLink */

/*!
 * \brief Fetch the next packet out of the receive ring buffer.
 *
 * \return Pointer to an allocated ::NETBUF. If there is no
 *         no data available, then the function returns a
 *         null pointer. If the NIC's buffer seems to be
 *         corrupted, a pointer to 0xFFFF is returned.
 */
static int NicGetPacket (NETBUF **nbp)
{
   int      rc = -1;
   uint8_t *buffer;
   uint16_t len;


   buffer = enet_RxFrameGet(&len);
   if (buffer != NULL)
   {
      /* We allocate a pbuf chain of pbufs from the pool. */
      *nbp = NutNetBufAlloc(0, NBAF_DATALINK + (2 & (NUTMEM_ALIGNMENT - 1)), len);
 
      /* Copy received frame from ethernet driver buffer to stack buffer */
      if (*nbp != NULL) 
      { 
         /* Copy packet data . */
         memcpy((*nbp)->nb_dl.vp, buffer, len);
         
         /* Return success. */
         rc = 0;
      }

      enet_RxFrameReady();
   }
   
   return(rc);
} /* NicGetPacket */

/*************************************************************************/
/*  NicWriteBuffer                                                       */
/*                                                                       */
/*  In    : buffer, src, sz                                              */
/*  Out   : none                                                         */
/*  Return: buffer                                                       */
/*************************************************************************/
static uint8_t *NicWriteBuffer (uint8_t *buffer, uint8_t *src, int sz)
{
   if (sz > 0)
   {
      memcpy(buffer, src, (size_t)sz),
      buffer += sz;
   }
   
   return(buffer);
} /* NicWriteBuffer */

/*************************************************************************/
/*  ReceiverTask                                                         */
/*                                                                       */
/*  Handle all incomming packets here.                                   */
/*                                                                       */
/*  In    : arg                                                          */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void ReceiverTask (void *arg)
{
   NUTDEVICE *dev = (NUTDEVICE*)arg;
   IFNET     *ifn;
   NICINFO   *ni;
   NETBUF    *nb;
   uint8_t   *buffer;
   uint32_t   linktime = OSTimeGet();
   unsigned   Event;
   
   ifn = (IFNET*)dev->dev_icb;
   ni  = (NICINFO*)dev->dev_dcb;

   /*
    * This is a temporary hack. Due to a change in initialization,
    * we may not have got a MAC address yet. Wait until a valid one
    * has been set.
    */
   while (ETHER_IS_ZERO(ifn->if_mac) || ETHER_IS_BROADCAST(ifn->if_mac)) 
   {
      OSTimeDly(10);
   }
   
   /* Set the MAC address in the MAC */ 
   enet_MACAddressConfig(ifn->if_mac); 

   /* Init the PHY */
   ni->ni_phy->Init(ni->ni_phy);

   /* Now the MAC can be started */
   enet_Start(); 

   while (1)
   {
      /* Wait for an event */
      Event = 0;
      OSEventWait(&NetIfEvent, NETIF_EVENT_WAIT_ALL, OS_EVENT_MODE_OR, &Event, 500); 
      
      if (Event & NETIF_EVENT_RX_OVERRUN)
      {
         enet_RxOverrun();
      }
      
      if (Event & NETIF_EVENT_RX)
      {
         /*
          * Fetch all packets from the NIC's internal buffer and pass
          * them to the registered handler.
          */
         while (NicGetPacket(&nb) == 0) 
         {
            /* Update monitoring counters. */
            NUT_PERFMON_ADD(ifn->if_in_octets, nb->nb_dl.sz);

            buffer = (uint8_t*)nb->nb_dl.vp;
            if (ETHER_IS_MULTICAST(buffer))
            {
               /* Multicast */
               NUT_PERFMON_INC(ifn->if_in_n_ucast_pkts);
            } 
            else 
            {
               NUT_PERFMON_INC(ifn->if_in_ucast_pkts);
            }
            
            /* Discard short packets. */
            if (nb->nb_dl.sz < 60) 
            {
               NutNetBufFree(nb);
               NUT_PERFMON_INC(ifn->if_in_errors);
            } 
            else 
            {
               (*ifn->if_recv)(dev, nb);
            }
            
            /* 
             * In case we have receive a packet, we can
             * restart the timeout for the CheckLink.
             */
            if (PHY_LINK_STATUS_NO_LINK == OldLinkStatus)
            {
               CheckLink(ni->ni_phy);
            }
            else
            {
               linktime = OSTimeGet();               
            }             
         }
      } /* if (Event & NETIF_EVENT_RX) */


      if (Event & NETIF_EVENT_TX_UNDERRUN)
      {
         enet_TxUnderrun();
      }
      
      /* Check timeout for CheckLink */
      if ((OSTimeGet() - linktime) > 1000)
      {
         linktime = OSTimeGet();
         CheckLink(ni->ni_phy);
      }
      
   } /* end while (1), task loop */
   
} /* ethernetif_input */

/*************************************************************************/
/*  devOutput                                                            */
/*                                                                       */
/*  Send Ethernet packet.                                                */
/*                                                                       */
/*  In    : dev, nb                                                      */
/*  Out   : none                                                         */
/*  Return: 0 on success, -1 in case of any errors.                      */
/*************************************************************************/
static int devOutput (NUTDEVICE *dev, NETBUF *nb)
{
   int      sz;
   IFNET   *nif = (IFNET*)dev->dev_icb;
   uint8_t *buffer;

   /* 
    * Calculate the number of bytes to be send. Do not send packets
    * larger than the Ethernet maximum transfer unit. The MTU consist
    * of 1500 data bytes plus the 14 byte Ethernet header plus 4 bytes
    * CRC. We check the data bytes only.
    */
   sz = nb->nb_nw.sz + nb->nb_tp.sz + nb->nb_ap.sz;
   if (sz > nif->if_mtu) 
   {
      return(-1);
   }
   
   /* Add the Ethernet header size. */
   sz += nb->nb_dl.sz;
   
   tal_CPUIrqDisable(ETHERNET_IRQn);

   buffer = enet_TxFrameGet();

   /* Transfer the Ethernet frame. */
   buffer = NicWriteBuffer(buffer, nb->nb_dl.vp, nb->nb_dl.sz);
   buffer = NicWriteBuffer(buffer, nb->nb_nw.vp, nb->nb_nw.sz);
   buffer = NicWriteBuffer(buffer, nb->nb_tp.vp, nb->nb_tp.sz);
            NicWriteBuffer(buffer, nb->nb_ap.vp, nb->nb_ap.sz);
   
   enet_TxFrameSend((uint16_t)sz);

   NUT_PERFMON_ADD(nif->if_out_octets, sz);

   tal_CPUIrqEnable(ETHERNET_IRQn);
   
   return(0);
} /* devOutput */

/*************************************************************************/
/*  devInit                                                              */
/*                                                                       */
/*  Initialize Ethernet hardware.                                        */
/*                                                                       */
/*  Resets the Ethernet controller, initializes all required hardware    */
/*  registers and starts a background thread for incoming Ethernet       */
/*  traffic.                                                             */
/*                                                                       */
/*  Applications should do not directly call this function.              */
/*  It is automatically executed during during device registration by    */
/*  NutRegisterDevice().                                                 */
/*                                                                       */
/*  If the network configuration hasn't been set by the application      */
/*  before registering the specified device, this function will          */
/*  call NutNetLoadConfig() to get the MAC address.                      */
/*                                                                       */
/*  In    : dev Identifies the device to initialize.                     */
/*  Out   : none                                                         */
/*  Return: 0 on success, -1 in case of any errors.                      */
/*************************************************************************/
static int devInit (NUTDEVICE *dev)
{
   NICINFO  *ni = (NICINFO*) dev->dev_dcb;

   /* Clear "Driver control block" which is our NICINFO */
   memset(ni, 0, sizeof(NICINFO));
   
   /* Configure the HW of the board, and return the PHY info */
   ni->ni_phy = BoardETHConfig();
   if (NULL == ni->ni_phy)
   {
      /* Error, no PHY available */
      return(-1);
   }

   /* Create the event used for RX and LinkStatus */
   OSEventCreate(&NetIfEvent);
   
   /* Start the receiver thread. */
   OSTaskCreate(&tcb_task_rx, ReceiverTask, dev, TASK_IP_RX_PRIORITY, 
                ReceiverStack, sizeof(ReceiverStack), 
                "iprx");
   
   return(0);
} /* devInit */

/*************************************************************************/
/*  devIOCtl                                                             */
/*                                                                       */
/*  In    : dev, req, conf.                                              */
/*  Out   : none                                                         */
/*  Return: 0 on success, -1 in case of any errors.                      */
/*************************************************************************/
static int devIOCtl (NUTDEVICE *dev, int req, void *conf)
{
   int         rc = 0;
   IFNET      *nif = (IFNET*) dev->dev_icb;
   NICINFO    *ni = (NICINFO*) dev->dev_dcb;
   uint32_t    ip_addr;
   MCASTENTRY *mcast;
   MCASTENTRY *mcast_prev;
   MCASTENTRY *mcast_next;
   uint8_t     mac[6];

   switch (req) 
   {
      case SIOCSIFADDR:    /* Set interface hardware address. */
         memcpy(nif->if_mac, conf, sizeof(nif->if_mac));
         memcpy(ni->ni_mac, conf, sizeof(ni->ni_mac));
         break;

      case SIOCADDMULTI:   /* Add multicast address. */
         /* Add multicast address. */
         ip_addr = *((uint32_t *) conf);
         rc = enet_MulticastAddFilter(ip_addr);
         if (0 == rc)
         {
            rc = -1;
            
            /* Test if the address is still in the list */
            mcast = nif->if_mcast;
            while (mcast)
            {
               if (ip_addr == mcast->mca_ip) 
               {
                  /* The address is still in the list, do nothing */
                  return(rc);
               }
               mcast = mcast->mca_next;
            }
            /* Create MAC address */
            mac[0] = 0x01;
            mac[1] = 0x00;
            mac[2] = 0x5E;
            mac[3] = ((uint8_t *) conf)[1] & 0x7f;
            mac[4] = ((uint8_t *) conf)[2];
            mac[5] = ((uint8_t *) conf)[3];
         
            mcast = malloc(sizeof(MCASTENTRY));
            if (mcast)
            {
               /* Add new mcast to the mcast list */
               memcpy(mcast->mca_ha, mac, 6);
               mcast->mca_ip = *((uint32_t *) conf);
               mcast->mca_next = nif->if_mcast;
               nif->if_mcast = mcast;
               
               rc = 0;
            }   
         }            
         break;
      
      case SIOCDELMULTI:   /* Delete multicast address. */
         /* Delete multicast address. */
         ip_addr = *((uint32_t *) conf);
         
         /* Remove from the list */   
         enet_MulticastDelFilter(ip_addr);
         
         /* Test if the address is still in the list */
         mcast = nif->if_mcast;
         mcast_prev = mcast;
         while (mcast) 
         {
            if (ip_addr == mcast->mca_ip) 
            {
               /* The address is in the list, leave the loop */
               break;
            }
            mcast_prev = mcast;
            mcast = mcast->mca_next;
         }
         if (NULL == mcast) 
         {
            /* The address is not in the list */
            return -1;
         }

         /*
          * Remove the address from the list
          */
         mcast_next = mcast->mca_next;

         /* Check if the first element must be removed */
         if (nif->if_mcast == mcast) 
         {
            /* The element is the first one. The first element is now the "next" element */
            nif->if_mcast = mcast_next;
            free(mcast);
         } 
         else 
         {
            /* The element is in the middle of the list. The next element of the previous is the "next" element */
            mcast_prev->mca_next = mcast_next;
            free(mcast);
         }
         break;
         
      default:
         rc = -1;
         break;
   }

   return(rc);
} /* devIOCTL */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  ETH_IRQHandler                                                       */
/*                                                                       */
/*  Ethernet IRQ handler                                                 */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void ENET_IRQHandler (void)
{
   uint32_t dStatus;
   
   TAL_CPU_IRQ_ENTER();

   /* Get the interrupt status */ 
   dStatus = enet_GetIRQStatus();
   
   /* 
    * RX interrupt handling 
    */
   if (dStatus & ENET_INT_RX_OVERRUN)
   {
      OSEventSetFromInt(&NetIfEvent, NETIF_EVENT_RX_OVERRUN);
   }    
   
   if ((dStatus & ENET_INT_RX) || (dStatus & ENET_INT_RX_ERROR))
   {
      OSEventSetFromInt(&NetIfEvent, NETIF_EVENT_RX);
   }      

   /*
    * TX interrupt handling 
    */
   if (dStatus & ENET_INT_TXUNDERRUN)
   {
      OSEventSetFromInt(&NetIfEvent, NETIF_EVENT_TX_UNDERRUN);
   } 
    
   if ((dStatus & ENET_INT_TX) || (dStatus & ENET_INT_TX_ERROR))
   {
      OSEventSetFromInt(&NetIfEvent, NETIF_EVENT_TX);
   }
   
   enet_ClearIRQStatus(dStatus);
   
   TAL_CPU_IRQ_EXIT();
} /* ETH_IRQHandler */


/*
 * Network interface information structure.
 */
static IFNET ifn_eth0 = 
{
   IFT_ETHER,              /* Interface type, if_type. */
   0,                      /* Interface flags, if_flags. */
   { 0, 0, 0, 0, 0, 0 },   /* Hardware net address, if_mac. */
   0,                      /* IP address, if_local_ip. */
   0,                      /* Remote IP address for point to point, if_remote_ip. */
   0,                      /* IP network mask, if_mask. */
   ETHERMTU,               /* Maximum size of a transmission unit, if_mtu. */
   0,                      /* Packet identifier, if_pkt_id. */
   0,                      /* Linked list of ARP entries, arpTable. */
   0,                      /* Linked list of multicast address entries, if_mcast. */
   NutEtherInput,          /* Routine to pass received data to, if_recv(). */
   devOutput,              /* Driver output routine, if_send(). */
   NutEtherOutput,         /* Media output routine, if_output(). */
   NULL                    /* Interface specific control function, if_ioctl(). */
#ifdef NUT_PERFMON
   , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
#endif
};


/*
 * Device information structure.
 *
 * A pointer to this structure must be passed to NutRegisterDevice()
 * to bind this Ethernet device driver to the Nut/OS kernel.
 * An application may then call NutNetIfConfig() with the name \em eth0
 * of this driver to initialize the network interface.
 */
NUTDEVICE devLPC17XX40XX = 
{
   0,                      /* Pointer to next device. */
   {'e', 't', 'h', '0', 0, 0, 0, 0, 0},   /* Unique device name. */
   IFTYP_NET,              /* Type of device. */
   0,                      /* Base address. */
   0,                      /* First interrupt number. */
   &ifn_eth0,              /* Interface control block. */
   &dcb_eth0,              /* Driver control block. */
   devInit,                /* Driver initialization routine. */
   devIOCtl,               /* Driver specific control function. */
   0,                      /* Read from device. */
   0,                      /* Write to device. */
   0,                      /* Open a device or file. */
   0,                      /* Close a device or file. */
   0,                      /* Request file size. */
   NULL,
   NULL
};

/*** EOF ***/
