/**************************************************************************
*  Copyright (c) 2014-2018 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  08.02.2014  mifi  First Version
*  15.02.2015  mifi  Some rework and cleanup.
*  28.04.2018  mifi  Simplified version. 
**************************************************************************/
#define __IPSTACK_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include "tal.h"
#include "ipstack.h"
#include "eth_phy_driver.h"

#include "sys/version.h"
#include "sys/thread.h"
#include "sys/timer.h"
#include "sys/device.h"
#include "sys/confnet.h"

#include "dev/dev_board.h"
#include "dev/urom.h"

/*=======================================================================*/
/*  Extern                                                               */
/*=======================================================================*/

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static NUTDEVICE *eth0_dev = NULL;

static uint8_t    MACDefaultAddr[6] = IP_DEFAULT_MAC_ADDR;
static uint8_t    MACAddr[6] = {0,0,0,0,0,0};
static uint32_t   IPAddr     = 0;
static uint32_t   NETMask    = 0;
static uint32_t   GWAddr     = 0;
static uint32_t   DNSAddr    = 0;

static int        LinkIsUp  = 0;

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  IPIfaceStart                                                         */
/*                                                                       */
/*  Start the TCP/IP stack.                                              */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: 0 on success / -1 otherwise                                  */
/*************************************************************************/
int IPIfaceStart (void)
{
   int rc = -1;
   
   /* Check if IP address or network mask is 0 */ 
   if ((0 == IPAddr) || (0 == NETMask))
   {
      /* Error, use default values */
      IPAddr  = inet_addr(IP_DEFAULT_ADDRESS);
      NETMask = inet_addr(IP_DEFAULT_NETMASK);
      GWAddr  = inet_addr(IP_DEFAULT_GATEWAY);
   }

   /* Check if DNS address is 0 */ 
   if (0 == DNSAddr)
   {
      DNSAddr = inet_addr(IP_DEFAULT_DNS);
   } 
   
   /* Check if MAC address is 0 */
   if( (0 == MACAddr[0]) && (0 == MACAddr[1]) && (0 == MACAddr[2]) &&
       (0 == MACAddr[3]) && (0 == MACAddr[4]) && (0 == MACAddr[5]) )
   {
      /* Error, use default values */
      memcpy(MACAddr, MACDefaultAddr, sizeof(MACAddr));   
   }
   
   /* Register the UROM file system */
   NutRegisterDevice(&devUrom, 0, 0);


   /*
    * Register Ethernet controller.
    */
   if (0 == NutRegisterDevice(&DEV_ETHER, 0, 0)) 
   {
      if (0 == NutNetIfConfig(DEV_ETHER_NAME, MACAddr, IPAddr, NETMask))
      {
         /* Without DHCP we had to set the default gateway manually.*/
         if(GWAddr) 
         {
            NutIpRouteAdd(0, 0, GWAddr, &DEV_ETHER);
         }

         eth0_dev = NutDeviceLookup("eth0");
         
         NutNetIfAddMcastAddr(DEV_ETHER_NAME, inet_addr("224.0.0.1"));
         
         rc = 0;
      }   
   } 
   
   return(rc);
} /* IPInit */

/*************************************************************************/
/*  IPIfaceIsReady                                                       */
/*                                                                       */
/*  Check if the link of the interface is up and an IP address is        */
/*  available.                                                           */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: 1 on success / 0 otherwise                                  */
/*************************************************************************/
int IPIfaceIsReady (void)
{
   int    IfIsReady = 0;
   IFNET *nif;

   /* Check if the device is valid */
   if (eth0_dev != NULL)
   {
      /* Get network interface */
      nif = eth0_dev->dev_icb;
   
      /* If link is up and ip is valid */
      if ((1 == LinkIsUp) && (nif->if_local_ip != 0))
      {
         IfIsReady = 1;
      }
   }      

   return(IfIsReady);
} /* IPIfaceIsReady */

/*************************************************************************/
/*  NetIfLinkCallback                                                    */
/*                                                                       */
/*  Callback is called when link is brought up/down                      */
/*                                                                       */
/*  In    : status                                                       */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void IPIfaceLinkCallback (uint8_t status)
{
   if (status != PHY_LINK_STATUS_NO_LINK)
   {
      if ((status & PHY_LINK_STATUS_SPEED_100M) && (status & PHY_LINK_STATUS_MODE_FULL))
      {
         TAL_PRINTF("*** eth0: Link is up: 100 MBit/s Full duplex ***\n");
      }   
      if ((status & PHY_LINK_STATUS_SPEED_100M) && (status & PHY_LINK_STATUS_MODE_HALF))
      {
         TAL_PRINTF("*** eth0: Link is up: 100 MBit/s Half duplex ***\n");
      }   
      if ((status & PHY_LINK_STATUS_SPEED_10M) && (status & PHY_LINK_STATUS_MODE_FULL))
      {
         TAL_PRINTF("*** eth0: Link is up: 10 MBit/s Full duplex ***\n");
      }   
      if ((status & PHY_LINK_STATUS_SPEED_10M) && (status & PHY_LINK_STATUS_MODE_HALF))
      {
         TAL_PRINTF("*** eth0: Link is up: 10 MBit/s Half duplex ***\n");
      }   
      
      LinkIsUp = 1;
   }
   else
   {
      TAL_PRINTF("*** eth0: Link is down ***\n");
      LinkIsUp = 0;
   }

   TAL_PRINTF("\n");
   
} /* IPIfaceLinkCallback */

/*** EOF ***/
